/*
* Theremin - Arduino Physical Computing fr Bastler, Designer & Geeks
*
* Soundcode von: Michael Smith <michael@hurts.ca> - http://www.arduino.cc/playground/Code/PCMAudio
* CapSense: http://www.arduino.cc/playground/Main/CapSense
*
* Alex Wenger 2009
*/
#include <CapSense.h>
#include <stdint.h>
#include <avr/interrupt.h>
#include <avr/io.h>
#include <avr/pgmspace.h>

// Audioausgabe
#define SAMPLE_RATE 8000

// Auf welchem PIN soll Audio ausgegeben werden? (nicht ndern!!!)
int speakerPin = 11;

// Sinustabelle fr Audiosignal
const unsigned char sintab[] PROGMEM = {  
  0x01,0x01,0x01,0x01,0x02,0x03,0x05,0x07,
  0x09,0x0c,0x0f,0x12,0x15,0x19,0x1c,0x21,
  0x25,0x29,0x2e,0x33,0x38,0x3d,0x43,0x48,
  0x4e,0x54,0x5a,0x60,0x66,0x6c,0x73,0x79,
  0x7f,0x85,0x8b,0x92,0x98,0x9e,0xa4,0xaa,
  0xb0,0xb6,0xbb,0xc1,0xc6,0xcb,0xd0,0xd5,
  0xd9,0xdd,0xe2,0xe5,0xe9,0xec,0xef,0xf2,
  0xf5,0xf7,0xf9,0xfb,0xfc,0xfd,0xfe,0xfe,
  0xfe,0xfe,0xfe,0xfd,0xfc,0xfb,0xf9,0xf7,
  0xf5,0xf2,0xef,0xec,0xe9,0xe5,0xe2,0xdd,
  0xd9,0xd5,0xd0,0xcb,0xc6,0xc1,0xbb,0xb6,
  0xb0,0xaa,0xa4,0x9e,0x98,0x92,0x8b,0x85,
  0x7f,0x79,0x73,0x6c,0x66,0x60,0x5a,0x54,
  0x4e,0x48,0x43,0x3d,0x38,0x33,0x2e,0x29,
  0x25,0x21,0x1c,0x19,0x15,0x12,0x0f,0x0c,
  0x09,0x07,0x05,0x03,0x02,0x01,0x01,0x01
};

volatile uint16_t sample;   // welches ist das nchste Sample aus der Sinustabelle
                            // die oberen 7 bit zeigen in die Tabelle, die restlichen 
                            // bits sind kommastellen
                            
uint16_t vol;               // aktuelle Lautstrke
volatile uint16_t set_vol;  // gewuenschte Lautstrke
volatile uint16_t tone;     // Tonhhe in "Inkrementiereinheiten"


// Interruptroutine, diese wird 8000 mal pro Sekunde aufgerufen und berechnet den nchsten
// Wert fr die Tonausgabe
ISR(TIMER1_COMPA_vect) {
  static int timer1counter;      // Zhler um Lautstrkenderung langsamer zu machen
  int wert;

  // Wert an der Stelle sample1/512 aus der sinus-Tabelle lesen
  wert = pgm_read_byte(&sintab[(sample >> 9)]);
  // Wert mit der aktuellen Lautstrke multiplizieren
  wert = (wert * vol) / 256; 
  // PWM Hardware anweisen ab jetzt diesen Wert auszugeben
  OCR2A = wert;

  // nchstes Sample in der Sinustabelle abhngig vom gewnschten
  // Ton auswhlen.
  sample += tone;                

  // Lautstrke anpassen wen gewnscht (nur alle 50 Interrupts, damit
  // es schn langsam passiert.
  timer1counter++;
  if (timer1counter > 50)
  {
    timer1counter = 0;  
    if (vol < set_vol) vol++;
    if (vol > set_vol) vol--;
  }
}

void startPlayback()
{
  pinMode(speakerPin, OUTPUT);

  // Initialisiere den Timer 2 fr die schnelle PWM zur
  // Soundausgabe auf Pin 11

  // Verwende den internen Takt (Datenblatt Seite 160)
  ASSR &= ~(_BV(EXCLK) | _BV(AS2));

  // Fast PWM mode  (Seite 157)
  TCCR2A |= _BV(WGM21) | _BV(WGM20);
  TCCR2B &= ~_BV(WGM22);

  // Whle die nicht invertierende PWM fr pin OC2A und OC2B
  // Am Arduino ist das Pin 11 und 3
  TCCR2A = (TCCR2A | _BV(COM2A1) | _BV(COM2B1));

  // Keine Vorteiler / wir wollen es schnell! (Seite 158)
  TCCR2B = (TCCR2B & ~(_BV(CS12) | _BV(CS11))) | _BV(CS10);

  // Start Wert = 0, sonst gibt es ein hssliches Ploppgerusch
  OCR2A = 0;
  OCR2B = 0;

  // Initialisiere Timer 1 fr 8000 Interrupts/Sekunde
  cli();

  // Set CTC mode (Clear Timer on Compare Match) (Seite 133)
  TCCR1B = (TCCR1B & ~_BV(WGM13)) | _BV(WGM12);
  TCCR1A = TCCR1A & ~(_BV(WGM11) | _BV(WGM10));

  // Kein Vorteiler (Seite 134)
  TCCR1B = (TCCR1B & ~(_BV(CS12) | _BV(CS11))) | _BV(CS10);

  // Gewnschte Frequenz = 8000kHz
  OCR1A = F_CPU / SAMPLE_RATE;    // 16e6 / 8000 = 2000

  // Aktiviere den Interrupt fr TCNT1 == OCR1A (Seite 136)
  TIMSK1 |= _BV(OCIE1A);

  // Startwerte
  sample = 0;

  // Global Interrupts wieder einschalten.
  sei();
}

// Aendert Ton und Lautstrke
// ton (50-4000Hz)
// volume (0-256);
void SetFreq(int ton, int volume)
{
  tone = (128ul*512ul*ton)/8000;
  set_vol  = volume;
}


CapSense   cs_4_2 = CapSense(4,2);        // 10M resistor between pins 4 & 2, pin 2 is sensor pin, add a wire and or foil if desired
CapSense   cs_4_6 = CapSense(4,6);        // 10M resistor between pins 4 & 6, pin 6 is sensor pin, add a wire and or foil

float volume_null = 100000;
float ton_null = 100000;

float gew_neg = 0.1;                      // Wie schnell sollen negative Werte weggerechnet werden?
float gew_pos = 0.001;                    // Wie schnell werden positive Werte weggerechnet / nicht zu gro whlen,
                                          // sonst verstimmt sich das Instrument beim spielen!

void Adjust_Null(float volume_raw, float ton_raw) {
  if (volume_raw < volume_null) {
    volume_null = volume_null * (1-gew_neg) + volume_raw * gew_neg;
  }
  if (volume_raw > volume_null) {
    volume_null = volume_null * (1-gew_pos) + volume_raw * gew_pos;
  }
  if (ton_raw < ton_null) {
    ton_null = ton_null * (1-gew_neg) + ton_raw * gew_neg;
  }
  if (ton_raw > ton_null) {
    ton_null = ton_null * (1-gew_pos) + ton_raw * gew_pos;
  }
}

float volume_max = 0;
float ton_max = 0;

void Adjust_Max(float volume_raw, float ton_raw) {
  if (volume_raw > volume_max) {
    volume_max = volume_max * (1-gew_neg) + volume_raw * gew_neg;
  }
  if (ton_raw > ton_max) {
    ton_max = ton_max * (1-gew_neg) + ton_raw * gew_neg;
  }
}

void setup()                    
{
  Serial.begin(9600);
  startPlayback();
}

int ton;
float ton_raw;
float volume_raw;
int volume;
int ton2;

void loop()                    
{
  long ton_tmp = cs_4_2.capSenseRaw(300);
  ton_raw    =  (ton_raw * 15 + ton_tmp)/16;
  volume_raw =  (volume_raw * 15 + cs_4_6.capSenseRaw(300))/16;
  
  Adjust_Null(volume_raw, ton_raw);
  Adjust_Max(volume_raw, ton_raw);
  
  ton    = map(ton_raw,ton_null,ton_max,50,880);
  ton    = constrain(ton,40,880);
  volume = map(volume_raw,volume_null,volume_max,0,1000);
  if (volume < 30) volume = 0;
  if (tone   < 50) volume = 0;
  volume = constrain(volume,0,255);
  
  SetFreq(ton,volume);
  
/*  Serial.print(ton);                  // print sensor output 1
  Serial.print("\t");
  Serial.println(volume);                  // print sensor output 2

  delay(10);                             // arbitrary delay to limit data to serial port 
  */
}
